@@ -3,18 +3,20 @@ from ._utils import (
     RandomStateLike,
     _cartesian_product,
     _check_array_key,
+    _check_compatible_fdata,
     _check_estimator,
+    _classifier_fit_depth_methods,
     _classifier_get_classes,
-    _domain_range,
+    _classifier_get_depth_methods,
     _evaluate_grid,
-    _FDataCallable,
     _int_to_real,
-    _pairwise_commutative,
+    _pairwise_symmetric,
     _reshape_eval_points,
     _same_domain,
     _to_array_maybe_ragged,
+    _to_domain_range,
     _to_grid,
-    _tuple_of_arrays,
+    _to_grid_points,
     check_is_univariate,
     nquad_vec,
 )
