@@ -2,14 +2,12 @@
 
 from abc import ABC
 
-from sklearn.base import BaseEstimator
-from sklearn.base import RegressorMixin
+import numpy as np
+from sklearn.base import BaseEstimator, RegressorMixin
 from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted
 
-import numpy as np
-
-from .. import FDataGrid, FData
-from ..misc.metrics import lp_distance
+from .. import FData, FDataGrid
+from ..misc.metrics import l2_distance
 
 
 def _to_multivariate(fdatagrid):
@@ -63,7 +61,7 @@ def _to_multivariate_metric(metric, grid_points):
 
         >>> import numpy as np
         >>> from skfda import FDataGrid
-        >>> from skfda.misc.metrics import lp_distance
+        >>> from skfda.misc.metrics import l2_distance
         >>> from skfda.ml._neighbors_base import _to_multivariate_metric
 
         Calculate the Lp distance between fd and fd2.
@@ -71,24 +69,24 @@ def _to_multivariate_metric(metric, grid_points):
         >>> x = np.linspace(0, 1, 101)
         >>> fd = FDataGrid([np.ones(len(x))], x)
         >>> fd2 =  FDataGrid([np.zeros(len(x))], x)
-        >>> lp_distance(fd, fd2).round(2)
+        >>> l2_distance(fd, fd2).round(2)
         array([ 1.])
 
         Creation of the sklearn-style metric.
 
-        >>> sklearn_lp_distance = _to_multivariate_metric(lp_distance, [x])
-        >>> sklearn_lp_distance(np.ones(len(x)), np.zeros(len(x))).round(2)
+        >>> sklearn_l2_distance = _to_multivariate_metric(l2_distance, [x])
+        >>> sklearn_l2_distance(np.ones(len(x)), np.zeros(len(x))).round(2)
         array([ 1.])
 
     """
     # Shape -> (n_samples = 1, domain_dims...., image_dimension (-1))
     shape = [1] + [len(axis) for axis in grid_points] + [-1]
 
-    def multivariate_metric(x, y, _check=False, **kwargs):
+    def multivariate_metric(x, y, **kwargs):
 
         return metric(_from_multivariate(x, grid_points, shape),
                       _from_multivariate(y, grid_points, shape),
-                      _check=_check, **kwargs)
+                      **kwargs)
 
     return multivariate_metric
 
@@ -162,7 +160,7 @@ class NeighborsMixin:
             if not self.multivariate_metric:
                 # Constructs sklearn metric to manage vector
                 if self.metric == 'l2':
-                    metric = lp_distance
+                    metric = l2_distance
                 else:
                     metric = self.metric
 
@@ -499,7 +497,7 @@ class NeighborsRegressorMixin(NeighborsMixin, RegressorMixin):
             if not self.multivariate_metric:
 
                 if self.metric == 'l2':
-                    metric = lp_distance
+                    metric = l2_distance
                 else:
                     metric = self.metric
 
