@@ -1,6 +1,10 @@
 """Classification."""
 from ._centroid_classifiers import DTMClassifier, NearestCentroid
-from ._depth_classifiers import MaximumDepthClassifier
+from ._depth_classifiers import (
+    DDClassifier,
+    DDGClassifier,
+    MaximumDepthClassifier,
+)
 from ._neighbors_classifiers import (
     KNeighborsClassifier,
     RadiusNeighborsClassifier,
